@@ -45,9 +45,16 @@ class Settings(BaseSettings):
         default=False,
         description="Use a cache to reduce processing times for evaluated prompts.",
     )
+    cache_size: int = Field(
+        default=2 << 30,
+        description="The size of the cache in bytes. Only used if cache is True.",
+    )
     vocab_only: bool = Field(
         default=False, description="Whether to only return the vocabulary."
     )
+    verbose: bool = Field(
+        default=True, description="Whether to print debug information."
+    )
 
 
 router = APIRouter()
@@ -83,9 +90,10 @@ def create_app(settings: Optional[Settings] = None):
         n_ctx=settings.n_ctx,
         last_n_tokens_size=settings.last_n_tokens_size,
         vocab_only=settings.vocab_only,
+        verbose=settings.verbose,
     )
     if settings.cache:
-        cache = llama_cpp.LlamaCache()
+        cache = llama_cpp.LlamaCache(capacity_bytes=settings.cache_size)
         llama.set_cache(cache)
     return app
 
@@ -349,7 +357,9 @@ GetModelResponse = create_model_from_typeddict(ModelList)
 
 
 @router.get("/v1/models", response_model=GetModelResponse)
-def get_models() -> ModelList:
+def get_models(
+    llama: llama_cpp.Llama = Depends(get_llama),
+) -> ModelList:
     return {
         "object": "list",
         "data": [
